#include "gtest_grail_backend.h"
#include "x11_mocks.h"

/*
  Check that GEIS_EVENT_ATTRIBUTE_CONSTRUCTION_FINISHED
  is properly filled.
 */

class ConstructionFinishedProperty : public Geis2GrailBackendBase
{
 protected:
  ConstructionFinishedProperty() : _subscription(nullptr),
                                   _gesture_update_count(0) {}

  void SendXInput2Events();

  virtual void OnEventInitComplete(GeisEvent event);
  virtual void OnEventGestureBegin(GeisEvent event);
  virtual void OnEventGestureUpdate(GeisEvent event);

  GeisSubscription _subscription;
  unsigned int _gesture_update_count;
};

void ConstructionFinishedProperty::SendXInput2Events()
{
  XEvent event;
  XGenericEventCookie *xcookie = 0;
  XIDeviceEvent *device_event = 0;
  XITouchOwnershipEvent *ownership_event = 0;
  int serial = 0;

  event.type = GenericEvent;
  device_event = (XIDeviceEvent*)calloc(1, sizeof(XIDeviceEvent));
  device_event->serial = serial++;
  device_event->display = xmock_display;
  device_event->extension = xmock_xi2_opcode;
  device_event->evtype = XI_TouchBegin;
  device_event->time = xmock_server_time;
  device_event->deviceid = 0;
  /* The source device that originally generated the event. */
  device_event->sourceid = device_event->deviceid;
  device_event->detail = 0; /* touch id */
  device_event->root = DefaultRootWindow(xmock_display);
  device_event->event = DefaultRootWindow(xmock_display);
  device_event->child = 0;
  device_event->root_x = 123.0f;
  device_event->root_y = 456.0f;
  device_event->event_x = device_event->root_x;
  device_event->event_y = device_event->root_y;
  device_event->valuators.mask_len = 2; /* two bytes */
  device_event->valuators.mask = (unsigned char*) malloc(2);
  XISetMask(device_event->valuators.mask, 0); /* X axis is present */
  XISetMask(device_event->valuators.mask, 1); /* Y axis is present */
  device_event->valuators.values = (double*) malloc(sizeof(double)*2);
  device_event->valuators.values[0] = -200.0;
  device_event->valuators.values[1] = -100.0;
  xcookie = &event.xcookie;
  xcookie->extension = xmock_xi2_opcode;
  xcookie->evtype = XI_TouchBegin;
  xcookie->data = device_event;
  xmock_add_to_event_queue(&event);

  event.type = GenericEvent;
  device_event = (XIDeviceEvent*)calloc(1, sizeof(XIDeviceEvent));
  device_event->serial = serial++;
  device_event->display = xmock_display;
  device_event->extension = xmock_xi2_opcode;
  device_event->evtype = XI_TouchBegin;
  device_event->time = xmock_server_time;
  device_event->deviceid = 0;
  device_event->sourceid = device_event->deviceid;
  device_event->detail = 1; /* touch id */
  device_event->root = DefaultRootWindow(xmock_display);
  device_event->event = DefaultRootWindow(xmock_display);
  device_event->child = 0;
  device_event->root_x = 222.0f;
  device_event->root_y = 456.0f;
  device_event->event_x = device_event->root_x;
  device_event->event_y = device_event->root_y;
  device_event->valuators.mask_len = 2;
  device_event->valuators.mask = (unsigned char*) malloc(2);
  XISetMask(device_event->valuators.mask, 0);
  XISetMask(device_event->valuators.mask, 1);
  device_event->valuators.values = (double*) malloc(sizeof(double)*2);
  device_event->valuators.values[0] = -150.0;
  device_event->valuators.values[1] = -100.0;
  xcookie = &event.xcookie;
  xcookie->extension = xmock_xi2_opcode;
  xcookie->evtype = XI_TouchBegin;
  xcookie->data = device_event;
  xmock_add_to_event_queue(&event);

  xmock_server_time += 5;

  event.type = GenericEvent;
  ownership_event = (XITouchOwnershipEvent*)calloc(1, sizeof(XITouchOwnershipEvent));
  ownership_event->type = GenericEvent;
  ownership_event->serial = serial++;
  ownership_event->display = xmock_display;
  ownership_event->extension = xmock_xi2_opcode;
  ownership_event->evtype = XI_TouchOwnership;
  ownership_event->time = xmock_server_time;
  ownership_event->deviceid = 0;
  ownership_event->sourceid = ownership_event->deviceid;
  ownership_event->touchid = 0;
  ownership_event->root = DefaultRootWindow(xmock_display);
  ownership_event->event = DefaultRootWindow(xmock_display);
  ownership_event->child = 0;
  xcookie = &event.xcookie;
  xcookie->extension = xmock_xi2_opcode;
  xcookie->evtype = XI_TouchOwnership;
  xcookie->data = ownership_event;
  xmock_add_to_event_queue(&event);

  event.type = GenericEvent;
  ownership_event = (XITouchOwnershipEvent*)calloc(1, sizeof(XITouchOwnershipEvent));
  ownership_event->type = GenericEvent;
  ownership_event->serial = serial++;
  ownership_event->display = xmock_display;
  ownership_event->extension = xmock_xi2_opcode;
  ownership_event->evtype = XI_TouchOwnership;
  ownership_event->time = xmock_server_time;
  ownership_event->deviceid = 0;
  ownership_event->sourceid = ownership_event->deviceid;
  ownership_event->touchid = 1;
  ownership_event->root = DefaultRootWindow(xmock_display);
  ownership_event->event = DefaultRootWindow(xmock_display);
  ownership_event->child = 0;
  xcookie = &event.xcookie;
  xcookie->extension = xmock_xi2_opcode;
  xcookie->evtype = XI_TouchOwnership;
  xcookie->data = ownership_event;
  xmock_add_to_event_queue(&event);

  event.type = GenericEvent;
  device_event = (XIDeviceEvent*)calloc(1, sizeof(XIDeviceEvent));
  device_event->serial = serial++;
  device_event->display = xmock_display;
  device_event->extension = xmock_xi2_opcode;
  device_event->evtype = XI_TouchUpdate;
  device_event->time = xmock_server_time;
  device_event->deviceid = 0;
  device_event->sourceid = device_event->deviceid;
  device_event->detail = 0;
  device_event->root = DefaultRootWindow(xmock_display);
  device_event->event = DefaultRootWindow(xmock_display);
  device_event->child = 0;
  device_event->root_x = 123.0f;
  device_event->root_y = 466.0f;
  device_event->event_x = device_event->root_x;
  device_event->event_y = device_event->root_y;
  device_event->valuators.mask_len = 2; /* two bytes */
  device_event->valuators.mask = (unsigned char*) malloc(2);
  XISetMask(device_event->valuators.mask, 0);
  XISetMask(device_event->valuators.mask, 1);
  device_event->valuators.values = (double*) malloc(sizeof(double)*2);
  device_event->valuators.values[0] = -200.0;
  device_event->valuators.values[1] = -90.0;
  xcookie = &event.xcookie;
  xcookie->extension = xmock_xi2_opcode;
  xcookie->evtype = XI_TouchUpdate;
  xcookie->data = device_event;
  xmock_add_to_event_queue(&event);

  event.type = GenericEvent;
  device_event = (XIDeviceEvent*)calloc(1, sizeof(XIDeviceEvent));
  device_event->serial = serial++;
  device_event->display = xmock_display;
  device_event->extension = xmock_xi2_opcode;
  device_event->evtype = XI_TouchUpdate;
  device_event->time = xmock_server_time;
  device_event->deviceid = 0;
  device_event->sourceid = device_event->deviceid;
  device_event->detail = 1; /* touch id */
  device_event->root = DefaultRootWindow(xmock_display);
  device_event->event = DefaultRootWindow(xmock_display);
  device_event->child = 0;
  device_event->root_x = 222.0f;
  device_event->root_y = 466.0f;
  device_event->event_x = device_event->root_x;
  device_event->event_y = device_event->root_y;
  device_event->valuators.mask_len = 2;
  device_event->valuators.mask = (unsigned char*) malloc(2);
  XISetMask(device_event->valuators.mask, 0);
  XISetMask(device_event->valuators.mask, 1);
  device_event->valuators.values = (double*) malloc(sizeof(double)*2);
  device_event->valuators.values[0] = -150.0;
  device_event->valuators.values[1] = -90.0;
  xcookie = &event.xcookie;
  xcookie->extension = xmock_xi2_opcode;
  xcookie->evtype = XI_TouchUpdate;
  xcookie->data = device_event;
  xmock_add_to_event_queue(&event);

  /* Go past composition time (i.e. time window where new touches can be grouped
     with existing ones to make new gestures) */
  xmock_server_time += 70;

  /* the next two updates (one for each touch point) should generate
     a gesture update from Grail whose slice will have its "construction
     finished" property set to true (because we're beyond composition time) */

  event.type = GenericEvent;
  device_event = (XIDeviceEvent*)calloc(1, sizeof(XIDeviceEvent));
  device_event->serial = serial++;
  device_event->display = xmock_display;
  device_event->extension = xmock_xi2_opcode;
  device_event->evtype = XI_TouchUpdate;
  device_event->time = xmock_server_time;
  device_event->deviceid = 0;
  device_event->sourceid = device_event->deviceid;
  device_event->detail = 0;
  device_event->root = DefaultRootWindow(xmock_display);
  device_event->event = DefaultRootWindow(xmock_display);
  device_event->child = 0;
  device_event->root_x = 123.0f;
  device_event->root_y = 476.0f;
  device_event->event_x = device_event->root_x;
  device_event->event_y = device_event->root_y;
  device_event->valuators.mask_len = 2; /* two bytes */
  device_event->valuators.mask = (unsigned char*) malloc(2);
  XISetMask(device_event->valuators.mask, 0);
  XISetMask(device_event->valuators.mask, 1);
  device_event->valuators.values = (double*) malloc(sizeof(double)*2);
  device_event->valuators.values[0] = -200.0;
  device_event->valuators.values[1] = -80.0;
  xcookie = &event.xcookie;
  xcookie->extension = xmock_xi2_opcode;
  xcookie->evtype = XI_TouchUpdate;
  xcookie->data = device_event;
  xmock_add_to_event_queue(&event);

  event.type = GenericEvent;
  device_event = (XIDeviceEvent*)calloc(1, sizeof(XIDeviceEvent));
  device_event->serial = serial++;
  device_event->display = xmock_display;
  device_event->extension = xmock_xi2_opcode;
  device_event->evtype = XI_TouchUpdate;
  device_event->time = xmock_server_time;
  device_event->deviceid = 0;
  device_event->sourceid = device_event->deviceid;
  device_event->detail = 1; /* touch id */
  device_event->root = DefaultRootWindow(xmock_display);
  device_event->event = DefaultRootWindow(xmock_display);
  device_event->child = 0;
  device_event->root_x = 222.0f;
  device_event->root_y = 476.0f;
  device_event->event_x = device_event->root_x;
  device_event->event_y = device_event->root_y;
  device_event->valuators.mask_len = 2;
  device_event->valuators.mask = (unsigned char*) malloc(2);
  XISetMask(device_event->valuators.mask, 0);
  XISetMask(device_event->valuators.mask, 1);
  device_event->valuators.values = (double*) malloc(sizeof(double)*2);
  device_event->valuators.values[0] = -150.0;
  device_event->valuators.values[1] = -80.0;
  xcookie = &event.xcookie;
  xcookie->extension = xmock_xi2_opcode;
  xcookie->evtype = XI_TouchUpdate;
  xcookie->data = device_event;
  xmock_add_to_event_queue(&event);
}

void ConstructionFinishedProperty::OnEventInitComplete(GeisEvent event)
{
  _subscription = CreateFilteredSubscription(
      "My 2-touches Touch", 2, GEIS_GESTURE_TOUCH);
  ASSERT_NE(nullptr, _subscription);

  SendXInput2Events();
}

void ConstructionFinishedProperty::OnEventGestureBegin(GeisEvent event)
{
  GeisAttr attr = geis_event_attr_by_name(event,
      GEIS_EVENT_ATTRIBUTE_CONSTRUCTION_FINISHED);

  GeisBoolean construction_finished = geis_attr_value_to_boolean(attr);

  ASSERT_EQ(GEIS_FALSE, construction_finished);
}

void ConstructionFinishedProperty::OnEventGestureUpdate(GeisEvent event)
{
  GeisAttr attr = geis_event_attr_by_name(event,
      GEIS_EVENT_ATTRIBUTE_CONSTRUCTION_FINISHED);

  GeisBoolean construction_finished = geis_attr_value_to_boolean(attr);

  if (_gesture_update_count == 0)
  {
    ASSERT_EQ(GEIS_FALSE, construction_finished);
  }
  else
  {
    /* on the second update we're past composition time and therefore grail will
       send a slice with "construction finished" set to true.
       geis must follow suit. */
    ASSERT_EQ(GEIS_TRUE, construction_finished);
  }

  ++_gesture_update_count;
}

TEST_F(ConstructionFinishedProperty, Test)
{
  CreateXMockTouchScreenDevice();

  Geis geis = geis_new(GEIS_INIT_UTOUCH_GRAIL_BACKEND,
                       GEIS_INIT_NO_ATOMIC_GESTURES,
                       nullptr);
  ASSERT_NE(nullptr, geis);

  Run(geis);

  /* there should have been two gesture update events from geis */
  ASSERT_EQ(2, _gesture_update_count);

  if (_subscription)
    geis_subscription_delete(_subscription);
  geis_delete(geis);

  DestroyXMockDevices();
}
